# Handles install routine - `amz install`
import subprocess as sp
import os
import shutil
from itertools import zip_longest

from amz.loading import Loading
import amz.meta as meta


# Helper func to install apt, py2, py3 deps
def std_deps(dep_type, filepath, verbose=False, prestr=''):

    # Validate type
    if dep_type not in ['apt', 'py2', 'py3']:
        print('Dependency type not among apt/py2/py3')
        print('Exiting...')
        return

    # list command
    list_cmd = {'apt': 'apt list --installed', 'py2': 'python2 -m pip list', 'py3': 'python3 -m pip list'}

    # Install command before pkgname
    install_cmd = {
        'apt': 'sudo apt --yes install',
        'py2': 'python2 -m pip install --user',
        'py3': 'python3 -m pip install --user'
    }

    filepath_display = filepath
    dep_list = []
    exist_list = []

    if isinstance(filepath, str):
        try:
            if os.path.exists(filepath):
                if len(filepath) > 30:
                    filepath_display = f'...{filepath[-25:]}'
            else:
                print(f'{filepath} - file does not exist')
                print('Exiting...')
                return
        except:
            return

        dep_list = open(filepath, 'r').read().split('\n')
        # Refine Lists not to contain comments(#) or empty lines
        dep_list = [pkg for pkg in dep_list if pkg != '' and '#' not in pkg]

    if isinstance(filepath, list):

        dep_list = filepath

    l = Loading(
        input_str=
        f"Installing {dep_type} dependencies {f'from - {filepath_display}' if isinstance(filepath, str) else ''}",
        input_prestr=prestr)
    l.end(-2)

    p0 = sp.run(list_cmd[dep_type], shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    exist_list = [x.split("/" if dep_type == 'apt' else " ")[0].lower() for x in p0.stdout.decode().split('\n')]

    for package in dep_list:
        if package.lower() in exist_list:
            if not verbose:
                l.chain(f'{package} already installed', prestr + '\t')
                l.end(0)
            elif verbose:
                print(f'{package} already installed')
        elif package.lower() not in exist_list:
            if not verbose: l.chain(f'{package}', prestr + '\t')
            p = sp.run(f'{install_cmd[dep_type]} {package}',
                       shell=True,
                       stdout=(None if verbose else sp.PIPE),
                       stderr=(None if verbose else sp.PIPE))
            if not verbose: l.end(p.returncode)


def script(filepath, input_args, verbose=False, forced=False, mode='install'):
    # Validate the script
    if not meta.validate_script(filepath):
        print(f'Invalid Installer - {filepath}')
        print('Exiting...')
        return
    # Load the script meta and install instructions
    script_meta = meta.get_yaml_config(filepath, ['meta'])
    script_install = meta.get_yaml_config(filepath, [mode])
    # Parse args
    args_exist = "args" in list(script_meta.keys())
    sub_map = []
    sub_map_key = []
    if args_exist:
        arg_dict = script_meta["args"]
        available_args = list(arg_dict.keys())
        if len(input_args) > len(available_args):
            print('More number of args given than possible')
            print('Exiting...')
            return
        # Create substitution map
        sub_map = [
            (f"[({key})]",
             arg_dict[key][arg] if arg is not None and arg in list(arg_dict[key].keys()) else arg_dict[key]['default'])
            for key, arg in zip_longest(available_args, input_args)
        ]
        sub_map_key = [(f"[({key}.key)]", arg if arg is not None and arg in list(arg_dict[key].keys()) else 'default')
                       for key, arg in zip_longest(available_args, input_args)]
    # Handle Temp Dir
    CUR_DIR = os.getcwd()
    TMP_DIR = f"{os.environ['HOME']}/.tmp-amz-{script_meta['name']}"
    if script_meta["temp_dir"] and mode in ['install', 'update']:
        # remove if it exists
        if os.path.exists(TMP_DIR):
            shutil.rmtree(TMP_DIR)
        os.mkdir(TMP_DIR)
        os.chdir(TMP_DIR)
    # Check if installed
    p0 = sp.run(script_meta["check"], shell=True, stderr=sp.PIPE, stdout=sp.PIPE)
    if p0.stdout.decode() == 'true\n' and not forced and mode == 'install':
        l = Loading(f"Already installed {script_meta['name']}")
        l.end(0)
        return
    if p0.stdout.decode() == 'false\n' and mode in ['update', 'remove']:
        l = Loading(f"{script_meta['name']} not installed to {mode}")
        l.end(0)
        return
    # Start Display
    pc_form = {'install': 'Installing', 'update': 'Updating', 'remove': 'Removing'}
    l = Loading(f"{'(FORCED) ' if forced else ''}{pc_form[mode]} {script_meta['name']}")
    l.end(-2)
    # Install deps
    if 'deps' in list(script_meta.keys()):
        dep_keys = list(script_meta['deps'].keys())
        for dep_type in dep_keys:
            std_deps(dep_type, script_meta['deps'][dep_type], verbose, '\t')
    # Install steps
    for section in list(script_install.keys()):
        # Check if section is blocking
        is_blocking = '!!!' in section
        disp_title = section.replace('!!!', '')
        ## Perform substitution for display title
        if args_exist:
            for key, val in sub_map + sub_map_key:
                disp_title = disp_title.replace(key, val)
        # Anim start
        if not verbose: l.chain(f'{disp_title}', '\t')
        # Final return code
        retcode = 0
        # Loop through commands under section
        for cmd in script_install[section]:
            ## Perform substitution for each commands
            if args_exist:
                for key, val in sub_map:
                    cmd = cmd.replace(key, val)
            # Execute via subprocess
            p = sp.run(cmd, shell=True, stdout=(None if verbose else sp.PIPE), stderr=(None if verbose else sp.PIPE))
            retcode += p.returncode
            # If blocking section, close if retcode is not 0
            if retcode != 0 and is_blocking:
                if not verbose: l.end(retcode)
                return
        # Anim end
        if not verbose: l.end(retcode)
    # Clean Up and Back to Cur Dir
    if script_meta["temp_dir"] and mode in ['install', 'update']:
        shutil.rmtree(TMP_DIR)
        os.chdir(CUR_DIR)



#!/usr/bin/env python3
'''
Assumes incoming files to be python or C/C++,
filtered from pre-commit framework
'''

import os
import sys
import re
from identify import identify

file_list = sys.argv[1:]

# Exit if no files given
if len(file_list) is 0:
    print("No files given!")
    sys.exit(0)


def check_file(filepath: str, comment_symbol: str):
    with open(filepath, 'r') as f:
        # Search for beginning of disclaimer
        for i, line in enumerate(f):
            line = line.replace(comment_symbol, '').strip()
            if 'AMZ Driverless Project' == line:
                break
            # Give up after 10 lines:
            if i == 10:
                return False
        # Empty line
        if next(f).replace(comment_symbol, '').strip():
            return False
        # Copyright line
        line = next(f).replace(comment_symbol, '').strip()
        if re.compile(r'Copyright \(c\) 20(1|2)\d(-20(1|2)\d)? Authors:').match(line) is None:
            return False
        # List of authors
        for i, line in enumerate(f):
            line = line.replace(comment_symbol, '').strip()
            # Reached end of the list
            if not line:
                if i == 0:
                    return False
                break
            if re.compile(r'- .+?(?=<)<.+?(?=@)@.+?(?=>)>').match(line) is None:
                return False
        # The next lines need to match literally
        if next(f).replace(comment_symbol, '').strip() != 'All rights reserved.':
            return False
        if next(f).replace(comment_symbol, '').strip():
            return False
        line = next(f).replace(comment_symbol, '').strip()
        if line != 'Unauthorized copying of this file, via any medium is strictly prohibited':
            return False
        if next(f).replace(comment_symbol, '').strip() != 'Proprietary and confidential':
            return False
    return True


failed_files = []
error_files = []

# Check all given files
for file in file_list:
    # Skip empty files like __init__.py
    if os.stat(file).st_size == 0:
        continue
    # Use tag indentification as being done in pre-commit
    tags = identify.tags_from_path(file)
    if 'python' in tags:
        if not check_file(file, '#'):
            failed_files.append(file)
    elif 'c++' in tags:
        if not check_file(file, '*'):
            failed_files.append(file)
    else:
        error_files.append(file)

COLLECTED_EXIT = 0
COLLECTED_OUTPUT = ''

# Print error messages
if failed_files:
    COLLECTED_EXIT += 1
    COLLECTED_OUTPUT += '\n----------\033[31m Did not find proper AMZ copyright disclaimers. \033[0m\n'
    COLLECTED_OUTPUT += '\nThe following files failed:\n'
    COLLECTED_OUTPUT += '\t' + '\n\t'.join(failed_files)
if error_files:
    COLLECTED_EXIT += 1
    COLLECTED_OUTPUT += '\n----------\033[31m The following files could not be indentified as Python or C/C++ code \033[0m\n'
    COLLECTED_OUTPUT += '\nThe following files failed:\n'
    COLLECTED_OUTPUT += '\t' + '\n\t'.join(error_files)

COLLECTED_OUTPUT += '\n----------'

if COLLECTED_EXIT != 0:
    print(COLLECTED_OUTPUT)
sys.exit(COLLECTED_EXIT)



# Helper for process handling - fancy wrapper around subprocess

import time, sys, os
import subprocess as sp

from amz.loading import Loading


# Classifies into normal commands, python commands
def classify_cmds(command):
    # Acquire prefix code
    try:
        cmd_prefix = command[:4]
    except IndexError:
        return -1, None
    # Check python command
    if cmd_prefix == 'pyx>':
        return 1, command[4:]
    # Check Re-sudo command
    elif cmd_prefix == 'rsu>':
        return 2, command[4:]
    # Else it's a normal command
    else:
        return 0, command


def run_cmd(command, name='', prestr='', verbose=False):

    l = Loading(name if name != '' else f'Executing `{command}`', input_prestr=prestr)
    cmd_type, cmd_exec = classify_cmds(command)

    if cmd_type == 0:
        p = sp.run(cmd_exec, shell=True, stdout=None if verbose else sp.PIPE, stderr=None if verbose else sp.PIPE)
        l.end(p.returncode)
    elif cmd_type == 1:
        try:
            exec(cmd_exec)
            l.end(0)
        except:
            l.end(1)
    elif cmd_type == 2:
        p = sp.run(f"sudo -p '{prestr}Need [sudo] access again: ' echo '{prestr}Thanks!'", shell=True)
        l.end(0)
    elif cmd_type == -1:
        print(f"Invalid Command - {command}")

    # Give some rest
    time.sleep(0.1)



#!/usr/bin/env python3
'''
Run appropriate linter on python files
Assumes incoming files to be python, 
filtered from pre-commit framework
'''

import os, sys
import subprocess as sp

file_list = sys.argv[1:]

# Exit if no files given
if len(file_list) is 0:
    print("No files given!")
    sys.exit(0)

py2_files = []
py3_files = []
err_files = []

# Classify files into python2, py3 and non-shebanged files
for file in file_list:
    # Skip empty files like __init__.py
    if os.stat(file).st_size == 0:
        continue
    with open(file) as f:
        fl = f.readlines()[0]
        if "python2" in fl:
            py2_files.append(file)
        elif "python3" in fl:
            py3_files.append(file)
        else:
            err_files.append(file)

COLLECTED_EXIT = 0
COLLECTED_OUTPUT = ''

# Lint Py2 files
if py2_files:
    p = sp.run(f"python2 -m pylint {' '.join(py2_files)}", shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    COLLECTED_EXIT += p.returncode
    if p.returncode != 0:
        COLLECTED_OUTPUT += '\n----------\033[31m Pylint on Py2 Files \033[0m\n'
        COLLECTED_OUTPUT += p.stdout.decode()

# Lint Py3 files
if py3_files:
    # for file in py3_files:
    p = sp.run(f"python3 -m pylint {' '.join(py3_files)}", shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    COLLECTED_EXIT += p.returncode
    if p.returncode != 0:
        COLLECTED_OUTPUT += '\n----------\033[31m Pylint on Py3 Files \033[0m\n'
        COLLECTED_OUTPUT += p.stdout.decode()

# Finally deal with non-shebanged code
if err_files:
    COLLECTED_EXIT += 1
    COLLECTED_OUTPUT += '\n----------\033[31m Some Error Files \033[0m\n'
    COLLECTED_OUTPUT += f"The following files possibly do not contain the right shebangs! \nDid not run pylint on:\n"
    COLLECTED_OUTPUT += '\t' + '\n\t'.join(err_files)
    COLLECTED_OUTPUT += '\n'

COLLECTED_OUTPUT += "----------"

if COLLECTED_EXIT != 0:
    print(COLLECTED_OUTPUT)
sys.exit(COLLECTED_EXIT)


# Handles fetching of meta data
import os
from functools import reduce
from operator import getitem
from itertools import product
import subprocess as sp

import yaml


def update_line_in_file(replace_containing, replace_with, filename):
    f = open(filename, 'r')
    lines = f.readlines()
    lines_upd = [f"{replace_with}\n" if replace_containing in l else l for l in lines]
    f.close()
    with open(filename, 'w') as f:
        f.write(''.join(lines_upd))
        f.close()


def get_version():
    PKG_DIR = os.path.dirname(__file__)
    v = open(f'{PKG_DIR}/data/VERSION').readlines()[0]
    return v.ljust(10, ' ')


def get_amz_config(key_chain):

    HOME_DIR = os.environ["HOME"]
    AMZ_CONFIG_FILE = HOME_DIR + '/.amz/config/amz-config.yaml'
    try:
        config = yaml.load(open(AMZ_CONFIG_FILE, 'r'), Loader=yaml.FullLoader)
        return reduce(getitem, key_chain, config)
    except:
        return


def get_yaml_config(filepath, key_chain):

    try:
        config = yaml.load(open(filepath, 'r'), Loader=yaml.FullLoader)
        return reduce(getitem, key_chain, config)
    except:
        return


def highlight_cmd(cmd):
    return f'\033[90m\033[107m {cmd} \033[0m'


def validate_script(filepath):
    # Open and parse yaml file
    yf = yaml.load(open(filepath), Loader=yaml.FullLoader)
    # Check for 4 sections - Meta, Install, Update, Remove
    if list(yf.keys()).sort() != ['meta', 'install', 'update', 'remove'].sort(): return False
    # Meta Section
    ## Reqd Keys
    if 'name' not in list(yf['meta'].keys()): return False
    if 'check' not in list(yf['meta'].keys()): return False
    ## deps keys are known?
    if 'deps' in list(yf['meta'].keys()):
        if not set(yf['meta']['deps'].keys()).issubset(set(['apt', 'py2', 'py3'])): return False
    ## args have default keys?
    if 'args' in list(yf['meta'].keys()):
        if not all('default' in val for val in list(yf['meta']['args'].values())): return False
    # Install/Update/Remove Section
    for sec in ['install', 'update', 'remove']:
        ## Each dict value is a list of str
        if not all(all(isinstance(cmd, str) for cmd in cmds) for cmds in list(yf[sec].values())): return False
    return True


def autocomplete_choices():
    try:
        amz_yaml = ''
        p = sp.run('git rev-parse --show-toplevel', shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
        # Check if your current repo is recognized as amz
        if p.returncode is 0 and os.path.exists(f'{p.stdout.decode()[:-1]}/amz.yaml'):
            amz_yaml = f'{p.stdout.decode()[:-1]}/amz.yaml'
        else:
            amz_yaml = get_amz_config(['amz_repo_meta', 'amz_repo_config'])
        # Fetch autocomplete options
        auto_yaml = get_yaml_config(amz_yaml, [])
        if auto_yaml is None:
            return None
        choices = list(set(['apt-deps', 'py2-deps', 'py3-deps']) & set(auto_yaml.keys()))
        if 'scripts' in auto_yaml.keys() and auto_yaml['scripts'] is not None:
            choices += list(auto_yaml['scripts'].keys())
            for scr in auto_yaml['scripts'].keys():
                args_avail = get_yaml_config(f'{p.stdout.decode()[:-1]}/{auto_yaml["scripts"][scr]}', ['meta', 'args'])
                if args_avail is not None:
                    arg_list = [list(args_avail[arg].keys())[1:] for arg in list(args_avail.keys())]
                    choices += [f'{scr}.' + '.'.join(opts) for opts in list(product(*arg_list))]
        return choices
    except:
        return None
